import 'package:dartz/dartz.dart';
import 'dart:convert';
import 'package:intl/intl.dart';

abstract class DataField {
  late bool _assigned;
  late dynamic _value;

  bool get isNull => !_assigned;
  bool get isNotNull => _assigned;

  dynamic get value {
    if (_assigned) {
      return some(_value);
    } else {
      return none();
    }
  }

  DataField() {
    _assigned = false;
  }

  void setValue(dynamic v);

  bool tryGetValue(bool Function(dynamic value) fn) {
    try {
      return fn(_value);
    } catch (e) {
      return false;
    }
  }

  void clear() {
    _value = null;
    _assigned = false;
  }

  String toJsonValue();
}

class IntegerDataField extends DataField {
  @override
  Option<int> get value {
    return _assigned ? some(_value as int) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is int) {
      _value = v;
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null), (a) => jsonEncode(a));
  }
}

class StringDataField extends DataField {
  @override
  Option<String> get value {
    return _assigned ? some(_value as String) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is String) {
      _value = v;
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null), (a) => jsonEncode(a));
  }
}

class NumericDataField extends DataField {
  @override
  Option<double> get value {
    return _assigned ? some(_value as double) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is double) {
      _value = v;
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null), (a) => jsonEncode(a));
  }
}

class BooleanDataField extends DataField {
  @override
  Option<bool> get value {
    return _assigned ? some(_value as bool) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is bool) {
      _value = v;
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null), (a) => jsonEncode(a));
  }
}

String format(String format, DateTime value) {
  return DateFormat(format).format(value);
}

class DateDataField extends DataField {
  @override
  Option<DateTime> get value {
    return _assigned ? some(_value as DateTime) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is DateTime) {
      _value = DateTime(v.year, v.month, v.day);
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(
        () => jsonEncode(null), (a) => jsonEncode(format('yyyy-MM-dd', a)));
  }
}

class DateTimeDataField extends DataField {
  @override
  Option<DateTime> get value {
    return _assigned ? some(_value as DateTime) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is DateTime) {
      _value = v;
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null),
        (a) => jsonEncode(format('yyyy-MM-dd HH:mm:ss', a)));
  }
}

class TimeDataField extends DataField {
  @override
  Option<DateTime> get value {
    return _assigned ? some(_value as DateTime) : none();
  }

  @override
  void setValue(dynamic v) {
    if (v is DateTime) {
      _value = DateTime(
          0, 0, 0, v.hour, v.minute, v.second, v.millisecond, v.microsecond);
      _assigned = true;
    }
  }

  @override
  String toJsonValue() {
    return value.fold(
        () => jsonEncode(null), (a) => jsonEncode(format('HH:mm:ss', a)));
  }
}

class DynamicDataField extends DataField {
  @override
  Option<dynamic> get value {
    return _assigned ? some(_value as dynamic) : none();
  }

  @override
  void setValue(dynamic v) {
    _value = v;
    _assigned = true;
  }

  @override
  String toJsonValue() {
    return value.fold(() => jsonEncode(null), (a) => jsonEncode(a));
  }
}

extension ExtendedMap on Map<String, DataField> {
  Option<DataField> getValue(String toKey) {
    var x;
    forEach((key, value) {
      if (key == toKey) {
        x = value;
      }
    });
    if (x != null) {
      return some(x);
    } else {
      return none();
    }
  }

  Option<int> valueAsInt(String key) {
    return getValue(key).fold(() => none(), (a) {
      if (a is IntegerDataField) {
        return a.value;
      } else {
        return none();
      }
    });
  }

  Option<String> valueAsString(String key) {
    return getValue(key).fold(() => none(), (a) {
      if (a is StringDataField) {
        return a.value;
      } else {
        return none();
      }
    });
  }

  Option<double> valueAsDouble(String key) {
    return getValue(key).fold(() => none(), (a) {
      if (a is NumericDataField) {
        return a.value;
      } else {
        return none();
      }
    });
  }

  Option<bool> valueAsBool(String key) {
    return getValue(key).fold(() => none(), (a) {
      if (a is BooleanDataField) {
        return a.value;
      } else {
        return none();
      }
    });
  }

  Option<DateTime> valueAsDateTime(String key) {
    return getValue(key).fold(() => none(), (a) {
      if (a is DateDataField) {
        return a.value;
      } else if (a is DateTimeDataField) {
        return a.value;
      } else if (a is TimeDataField) {
        return a.value;
      } else {
        return none();
      }
    });
  }
}

class Dataset {
  late Map<String, Type> _structure;
  int _rowIndex = -1;
  late List<Map<String, DataField>> _fields;

  Map<String, Type> get structure => _structure;

  int get count => _fields.length;

  bool get isEmpty => count == 0;
  bool get isNotempty => !isEmpty;

  Map<String, DataField> get fields {
    if (count > 0) {
      return _fields[_rowIndex];
    } else {
      return {};
    }
  }

  Dataset() {
    _fields = [];
    _structure = {};
  }

  void clearFields() {
    _rowIndex = -1;
    _fields.clear();
  }

  void clearStructure() {
    _structure.clear();
  }

  void clear() {
    clearFields();
    clearStructure();
  }

  bool first() {
    if (isNotempty) {
      _rowIndex = 0;
      return true;
    } else {
      return false;
    }
  }

  bool last() {
    if (isNotempty) {
      _rowIndex = count - 1;
      return true;
    } else {
      return false;
    }
  }

  bool next() {
    if (isNotempty) {
      if ((_rowIndex + 1) < count) {
        _rowIndex++;
        return true;
      } else {
        return false;
      }
    } else {
      return false;
    }
  }

  bool previous() {
    if (isNotempty) {
      if (_rowIndex > 0) {
        _rowIndex--;
        return true;
      } else {
        return false;
      }
    } else {
      return false;
    }
  }

  void forEach(void Function(Map<String, DataField> record) action) {
    for (var record in _fields) {
      action(record);
    }
  }

  Option<DataField> fieldByName(String name) {
    var x = fields;
    var v = x[name];
    return v != null ? some(v) : none();
  }

  bool trySetValue(String fieldName, dynamic value) {
    return fieldByName(fieldName).fold(() => false, (a) {
      a.setValue(value);
      return a._assigned;
    });
  }

  bool tryGetValue(String fieldName, bool Function(dynamic value) fn) {
    return fieldByName(fieldName).fold(() {
      return false;
    }, (a) {
      return a.tryGetValue(fn);
    });
  }

  void newRecord() {
    // ignore: omit_local_variable_types
    Map<String, DataField> x = {};

    _structure.forEach((key, value) {
      if (value.hashCode == (BooleanDataField).hashCode) {
        x.addAll({key: BooleanDataField()});
      } else if (value.hashCode == (IntegerDataField).hashCode) {
        x.addAll({key: IntegerDataField()});
      } else if (value.hashCode == (StringDataField).hashCode) {
        x.addAll({key: StringDataField()});
      } else if (value.hashCode == (NumericDataField).hashCode) {
        x.addAll({key: NumericDataField()});
      } else if (value.hashCode == (DateDataField).hashCode) {
        x.addAll({key: DateDataField()});
      } else if (value.hashCode == (DateTimeDataField).hashCode) {
        x.addAll({key: DateTimeDataField()});
      } else if (value.hashCode == (TimeDataField).hashCode) {
        x.addAll({key: TimeDataField()});
      } else {
        x.addAll({key: DynamicDataField()});
      }
    });

    _fields.add(x);
    _rowIndex++;
  }

  void deleteRecord() {
    if (isNotempty) {
      // ignore: unnecessary_null_comparison
      try {
        _fields.removeAt(_rowIndex);
        previous();
      } finally {}
      ;
    }
  }

  String toJson() {
    var x = '';
    var y = '';

    _fields.forEach((element) {
      x += ', {';
      y = '';
      element.forEach((key, value) {
        y += ', ${jsonEncode(key)}: ${value.toJsonValue()}';
      });
      x += y.replaceFirst(', ', '') + '}';
    });

    x = x.replaceFirst(', ', '');
    return '[$x]';
  }
}

String getJsonValue(String name, dynamic value) {
  return '${jsonEncode(name)}: ${jsonEncode(value)}';
}
